// CDC for xmega-prozessor

#include <util/atomic.h>
#include <avr/io.h>
#include "usb_defaults.h"
#include "usb_xm.h"
#include "usb_cdc.h"


volatile unsigned char usb_stored_control=0;

EP_buffer cdc_rxb;
EP_buffer cdc_txb;

///////////////////////////////////////////////////////////////////////////////////////
// define serial defaults
///////////////////////////////////////////////////////////////////////////////////////
CDC_LineEncoding_t LineEncoding = 
{ 
	.BaudRateBPS = 9600,
	.CharFormat  = CDC_LINEENCODING_OneStopBit,
	.ParityType  = CDC_PARITY_None,
	.DataBits    = 8
};


///////////////////////////////////////////////////////////////////////////////////////
// initialize the usb_cdc
///////////////////////////////////////////////////////////////////////////////////////
void usb_init(void)
{
	cdc_rxb.flag=0;
	cdc_rxb.len=0;
	cdc_rxb.bytes=0;

	cdc_txb.flag=0;
	cdc_txb.len=0;
	cdc_txb.bytes=0;

	epTX_ptr_in->DATAPTR=(unsigned int)epTX_buf_in;
	epTX_ptr_out->DATAPTR=(unsigned int)epTX_buf_out;
	epRX_ptr_in->DATAPTR=(unsigned int)epRX_buf_in;
	epRX_ptr_out->DATAPTR=(unsigned int)epRX_buf_out;
	epNOTE_ptr_in->DATAPTR=(unsigned int)epNOTE_buf_in;
	epNOTE_ptr_out->DATAPTR=(unsigned int)epNOTE_buf_out;

//	EP_DEF_init_buffer(ep_in);
//	EP_DEF_init_buffer(ep_out);
//	EP_DEF_init_buffer(ep_note);

	// Enable USB interrupts
//	USB.INTCTRLA = /*USB_SOFIE_bm |*/ USB_BUSEVIE_bm | USB_INTLVL_MED_gc;
//	USB.INTCTRLA = /*USB_SOFIE_bm |*/ USB_BUSEVIE_bm | USB_INTLVL_HI_gc;
//	USB.INTCTRLB = USB_TRNIE_bm | USB_SETUPIE_bm;

	USB_xm_Init();

}

///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
// Events 
///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////
// if Configuration is changed this is set
///////////////////////////////////////////////////////////////////////////////////////
void EVENT_USB_Device_ConfigurationChanged(uint8_t config)
{
	// set TX (endpoint in)
	epTX_bank=0;
	epTX_ptr_in->CTRL=0;
	epTX_ptr_in->CNT=0;
	epTX_ptr_in->STATUS=USB_EP_BUSNACK0_bm | USB_EP_BUSNACK1_bm;
	epTX_ptr_in->CTRL=USB_EP_TYPE_BULK_gc| USB_EP_BUFSIZE_64_gc | USB_EP_PINGPONG_bm;
	epTX_ptr_out->CNT=0;
	epTX_ptr_out->CTRL=0;
	epTX_ptr_out->STATUS=USB_EP_BUSNACK0_bm;

	// set NOTE (endpoint in)
	epNOTE_bank=0;
	epNOTE_ptr_in->CTRL=0;
	epNOTE_ptr_in->CNT=0;
	epNOTE_ptr_in->STATUS=USB_EP_BUSNACK0_bm | USB_EP_BUSNACK1_bm;
	epNOTE_ptr_in->CTRL=USB_EP_TYPE_BULK_gc| USB_EP_BUFSIZE_8_gc | USB_EP_PINGPONG_bm;
	epNOTE_ptr_out->CNT=0;
	epNOTE_ptr_out->CTRL=0;
	epNOTE_ptr_out->STATUS=USB_EP_BUSNACK0_bm;

	// set RX (endpoint out... incomming data)
	epRX_bank=0;
	epRX_ptr_out->CTRL=0;
	epRX_ptr_out->CNT=0;
	epRX_ptr_out->STATUS=0;
	epRX_ptr_out->CTRL=USB_EP_TYPE_BULK_gc| USB_EP_BUFSIZE_64_gc | USB_EP_PINGPONG_bm;
	epRX_ptr_in->CNT=0;
	epRX_ptr_in->CTRL=0;
	epRX_ptr_in->STATUS=0;
}


///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
void EVENT_USB_Device_ControlOUT(uint8_t* data, uint8_t len)
{
	if(len)
	{
		switch(usb_stored_control)
		{
			case CDC_REQ_SetLineEncoding:
				/* Read the line coding data in from the host into the global struct */
				memcpy(&LineEncoding,data,sizeof(CDC_LineEncoding_t));
				break; 
		}
	}
}

///////////////////////////////////////////////////////////////////////////////////////
// evnt for Control Request
///////////////////////////////////////////////////////////////////////////////////////
/** Event handler for the library USB Control Request reception event. */
bool EVENT_USB_Device_ControlRequest(USB_Request_Header_t* req)
{
	USB_EP_pair_t* pair = &endpoints[0];
	USB_EP_t* e = &pair->ep[1];
	USB_EP_t* b = &pair->ep[1];

	/* Process CDC specific control requests */
	switch (req->bRequest)
	{
		case CDC_REQ_GetLineEncoding:
			if (req->bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
			{

				/* Write the line coding data to the control endpoint */
				memcpy(ep0_buf_in,(char *)&LineEncoding,sizeof(CDC_LineEncoding_t));
				// send ep 0
				b->DATAPTR = (unsigned) ep0_buf_in;
				b->CNT = sizeof(CDC_LineEncoding_t);
				LACR16(&(e->STATUS), USB_EP_BUSNACK0_bm | USB_EP_TRNCOMPL0_bm);
				return true;
			}
			break;

		case CDC_REQ_SetLineEncoding:
			if (req->bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				usb_stored_control=CDC_REQ_SetLineEncoding;
				/* Read the line coding data in from the host into the global struct */
				// send ep 0
				b->DATAPTR = (unsigned) ep0_buf_in;
				b->CNT = 0;
				LACR16(&(e->STATUS), USB_EP_BUSNACK0_bm | USB_EP_TRNCOMPL0_bm);
				return true;
			}

			break;
		case CDC_REQ_SetControlLineState:
			if (req->bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
				// send ep 0
				b->DATAPTR = (unsigned) ep0_buf_in;
				b->CNT = 0;
				LACR16(&(e->STATUS), USB_EP_BUSNACK0_bm | USB_EP_TRNCOMPL0_bm);

				return true;
			}
			break;

		case CDC_REQ_SendBreak:
			if (req->bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
			{
//				EVENT_CDC_Device_BreakSent(CDCInterfaceInfo, (uint8_t)USB_ControlRequest.wValue);
				// send ep 0
				b->DATAPTR = (unsigned) ep0_buf_in;
				b->CNT = 0;
				LACR16(&(e->STATUS), USB_EP_BUSNACK0_bm | USB_EP_TRNCOMPL0_bm);
				return true;
			}

			break;
	}

//	if ((req->bmRequestType & CONTROL_REQTYPE_TYPE) == REQTYPE_VENDOR)
//	{
//		if (req->bRequest == 0xBB)
//		{
//			USB_enter_bootloader();
//		}
//	}
	return false;
}


