#ifndef _USB_XM_H
#define _USB_XM_H
#include "usb_defaults.h"
#include "Descriptors.h"

// Callback for incomming data from usb (host to device). 
// this functions have to handle the databuffer pointet with the void-ptr
// the data-len is given in le;
// return-value not used (return 0)

// callback for outgoing data to usb (device to host). 
// this functions have to handle the databuffer pointet with the void-ptr
// the data-max-len is given in le; it have to returnd the len of data in buffer
// if nothing to send, return 0

#define USB_EP_size_to_gc(x)  ((x <= 8   )?USB_EP_BUFSIZE_8_gc:\
                               (x <= 16  )?USB_EP_BUFSIZE_16_gc:\
                               (x <= 32  )?USB_EP_BUFSIZE_32_gc:\
                               (x <= 64  )?USB_EP_BUFSIZE_64_gc:\
                               (x <= 128 )?USB_EP_BUFSIZE_128_gc:\
                               (x <= 256 )?USB_EP_BUFSIZE_256_gc:\
                               (x <= 512 )?USB_EP_BUFSIZE_512_gc:\
                                           USB_EP_BUFSIZE_1023_gc)

#define USB_EP_IN 0x80

// Flag in the endpoint address to indicate that the endpoint should use
// PingPong (double buffer) mode. This is not actually part of the endpoint
// address as seen by the host. If PP is enabled, this flag needs to be part
// of the address passed to all USB_EP_* functions.
#define USB_EP_PP 0x40

extern USB_Request_Header_t USB_ControlRequest;

typedef union USB_EP_pair
{
	union
	{
		struct
		{
			USB_EP_t out;
			USB_EP_t in;
		};
		USB_EP_t ep[2];
	};
} ATTR_PACKED USB_EP_pair_t;

#define EP_NOTE_IDLE				0
#define EP_NOTE_SET_DCD				1
#define EP_NOTE_CLR_DCD				2
#define EP_NOTE_SEND				0x80


extern volatile uint8_t USB_DeviceState;
extern volatile uint8_t USB_Device_ConfigurationNumber;

extern uint8_t ep0_buf_in[USB_DEF_EP0_SIZE];
extern uint8_t ep0_buf_out[USB_DEF_EP0_SIZE];
extern USB_EP_pair_t endpoints[USB_DEF_EP_MAX];

extern uint8_t epTX_buf_in[CDC_TXRX_EPSIZE];
extern uint8_t epTX_buf_out[CDC_TXRX_EPSIZE];
extern uint8_t epTX_bank;
extern USB_EP_t *epTX_ptr_in;
extern USB_EP_t *epTX_ptr_out;

extern uint8_t epRX_buf_in[CDC_TXRX_EPSIZE];
extern uint8_t epRX_buf_out[CDC_TXRX_EPSIZE];
extern uint8_t epRX_bank;
extern USB_EP_t *epRX_ptr_in;
extern USB_EP_t *epRX_ptr_out;

extern uint8_t epNOTE_buf_in[CDC_NOTIFICATION_EPSIZE];
extern uint8_t epNOTE_buf_out[CDC_NOTIFICATION_EPSIZE];
extern uint8_t epNOTE_bank;
extern USB_EP_t *epNOTE_ptr_in;
extern USB_EP_t *epNOTE_ptr_out;

void USB_ep0_send(const uint8_t* addr, uint16_t size);
void USB_ResetInterface(void);
void USB_Evt_Task(void);
void USB_Task(void);
void USB_xm_Init(void);

void USB_T(void);
void USB_IO(void);

#endif
