/************************************************************************/
/* XBoot Extensible AVR Bootloader                                      */
/*                                                                      */
/* tested with ATXMEGA64A3, ATXMEGA128A1, ATXMEGA256A1, ATXMEGA32A4     */
/*                                                                      */
/* xboot.c                                                              */
/*                                                                      */
/* Alex Forencich <alex@alexforencich.com>                              */
/*                                                                      */
/* Copyright (c) 2010 Alex Forencich                                    */
/*                                                                      */
/* Permission is hereby granted, free of charge, to any person          */
/* obtaining a copy of this software and associated documentation       */
/* files(the "Software"), to deal in the Software without restriction,  */
/* including without limitation the rights to use, copy, modify, merge, */
/* publish, distribute, sublicense, and/or sell copies of the Software, */
/* and to permit persons to whom the Software is furnished to do so,    */
/* subject to the following conditions:                                 */
/*                                                                      */
/* The above copyright notice and this permission notice shall be       */
/* included in all copies or substantial portions of the Software.      */
/*                                                                      */
/* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,      */
/* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF   */
/* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND                */
/* NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS  */
/* BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN   */
/* ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN    */
/* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE     */
/* SOFTWARE.                                                            */
/*                                                                      */
/************************************************************************/

#include "usb_cdc.h"
#include "xboot.h"

void reset(void) __attribute__((naked,section(".vectors"))); 
void reset(void)
{ 
     asm("clr r1"); 
     SP=RAMEND; 
     SREG=0; 
     asm("jmp __dtors_end"); 
} 
 
void jmp_main(void) __attribute__((naked,section(".init9"))); 
void jmp_main(void)
{ 
     asm("jmp main"); 
} 

// Main code
int main(void)
{
	ADDR_T address = 0;
	unsigned char tmp;
	unsigned char in_bootloader;
	unsigned char val = 0;
#ifdef USE_WATCHDOG
	uint8_t temp;
#endif
	int i=0;
	void (*reset_vect)( void ) = 0x000000;
        

#ifdef USE_ENTER_PIN
	// Make sure it's an input
	ENTER_PORT.DIRCLR = (1 << ENTER_PIN);
	#if ENTER_PIN_PUEN
	// Enable bootloader entry pin pullup
	ENTER_PIN_CTRL = 0x18;
	#endif
#endif
        
	// Initialization section
	// Entry point and communication methods are initialized here
	// --------------------------------------------------


	/////////////////////////////////////////////////////////////////
	// get USBRCOSC
	NVM.CMD  = NVM_CMD_READ_CALIB_ROW_gc;
	tmp = pgm_read_byte(offsetof(NVM_PROD_SIGNATURES_t, USBRCOSC));
	/* Clean up NVM Command register. */ 
	NVM_CMD = NVM_CMD_NO_OPERATION_gc; 
	//
	/////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////
	// enable DFLL for 32MHz osz and trim to 48MHz sync with USB start of frame
	OSC.DFLLCTRL = OSC_RC32MCREF_USBSOF_gc;
	DFLLRC32M.CALB = tmp;
	DFLLRC32M.COMP1 = 0x1B; //Xmega AU manual, 4.17.19
	DFLLRC32M.COMP2 = 0xB7;
	DFLLRC32M.CTRL = DFLL_ENABLE_bm;
	//
	/////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////
	// enable 32 MHZ osz (trimmed to 48MHZ for usb)
	CCP = CCP_IOREG_gc; //Security Signature to modify clock 
	OSC.CTRL = OSC_RC32MEN_bm | OSC_RC2MEN_bm; // enable internal 32MHz oscillator and 2MHz
    
	while(!(OSC.STATUS & OSC_RC32MRDY_bm)); // wait for oscillator ready
	//
	/////////////////////////////////////////////////////////////////
	
	/////////////////////////////////////////////////////////////////
	// enable pll *16
	OSC.PLLCTRL = OSC_PLLSRC_RC2M_gc | 16; // 2MHz * 16 = 32MHz
	//
	/////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////
	// enable PLL
	CCP = CCP_IOREG_gc;
	OSC.CTRL = OSC_RC32MEN_bm | OSC_PLLEN_bm | OSC_RC2MEN_bm ; // Enable PLL
    
	while(!(OSC.STATUS & OSC_PLLRDY_bm)); // wait for PLL ready
	DFLLRC2M.CTRL = DFLL_ENABLE_bm;

	//
	/////////////////////////////////////////////////////////////////

	/////////////////////////////////////////////////////////////////
	// use PLL as systemclk
	CCP = CCP_IOREG_gc; /* allow changing CLK.CTRL */
	CLK.CTRL = CLK_SCLKSEL_PLL_gc; // use PLL output as system clock 
	//
	/////////////////////////////////////////////////////////////////
	CLK.PSCTRL = 0x00; // No peripheral clock prescaler

#ifdef USE_LED
	// Initialize LED pin
	LED_PORT.DIRSET = (1 << LED_PIN);
	#if LED_PIN_INV
		LED_PORT.OUTCLR = (1 << LED_PIN);
	#else
		LED_PORT.OUTSET = (1 << LED_PIN);
	#endif // LED_PIN_INV
#endif // USE_LED


        
	// --------------------------------------------------
	// End initialization section
	
	// One time trigger section
	// Triggers that are checked once, regardless of
	// whether or not USE_ENTER_DELAY is selected
	// --------------------------------------------------
        
	// --------------------------------------------------
	// End one time trigger section

	in_bootloader = 0;
        
#ifdef USE_WATCHDOG
	temp = WDT_ENABLE_bm | WDT_CEN_bm | WATCHDOG_TIMEOUT;
	CCP = CCP_IOREG_gc;
	WDT.CTRL = temp;
        
	/* Wait for WD to synchronize with new settings. */
	while(WDT_IsSyncBusy());

	// WDT_EnableAndSetTimeout();
#endif // USE_WATCHDOG

	if((ENTER_PORT.IN & (1 << ENTER_PIN)) == (ENTER_PIN_STATE ? (1 << ENTER_PIN) : 0))
	{
		// Check entry pin state
		in_bootloader = 1;
		// Initialize USB
		usb_init();
		// sign no data in inbuffer; get new one
		cdc_rxb.flag=1;
	}

	// Main bootloader        
	while (in_bootloader) 
	{

#ifdef USE_LED
		LED_PORT.OUTTGL = (1 << LED_PIN);
#endif // USE_LED
                
		val = get_char();

#ifdef USE_WATCHDOG
		WDT_Reset();
#endif // USE_WATCHDOG

		// Main bootloader parser
		// check autoincrement status
		if (val == 'a')
		{
			// yes, it is supported
			send_char('Y');
		}
		// Set address
		else if (val == 'A')
		{
			// Read address high then low
			address = (((ADDR_T)get_char() << 8) | (ADDR_T)get_char()) & 0x0000FFFF;
			// acknowledge
			send_char('\r');
		}
		// Extended address
		else if (val == 'H')
		{
			// Read address high then low
			address = (((ADDR_T)get_char() << 16) | ((ADDR_T)get_char() << 8) | (ADDR_T)get_char()) & 0x00FFFFFF;
			// acknowledge
			send_char('\r');
		}
		// Chip erase
		else if (val == 'e')
		{
			for (address = 0; address < APP_SECTION_SIZE; address += APP_SECTION_PAGE_SIZE)
			{
#ifdef USE_WATCHDOG
				WDT_Reset();
#endif // USE_WATCHDOG
				// wait for SPM instruction to complete
				SP_WaitForSPM();
				// erase page
				SP_EraseApplicationPage(address);
			}

			// Randomize page buffer
			EEPROM_LoadPage(&val);
			// Erase EEPROM
			EEPROM_EraseAll();

			// acknowledge
			send_char('\r');
		}
#ifdef ENABLE_BLOCK_SUPPORT
		// Check block load support
		else if (val == 'b')
		{
			// yes, it is supported
			send_char('Y');
			// Send block size (page size)
			send_char((APP_SECTION_PAGE_SIZE >> 8) & 0xFF);
			send_char(APP_SECTION_PAGE_SIZE & 0xFF);
		}
		// Block load
		else if (val == 'B')
		{
			unsigned int data;

			// Block size
			i = (get_char() << 8) | get_char();
			// Memory type
			val = get_char();

			// support eeprom
			if(val=='E')
			{
				EEPROM_FlushBuffer();
				// disable mapping of EEPROM into data space (enable IO mapped access)
				EEPROM_DisableMapping();
				do
				{
#ifdef USE_WATCHDOG
					WDT_Reset();
#endif // USE_WATCHDOG
					EEPROM_WaitForNVM();
					EEPROM_WriteByte( (unsigned char)(address / EEPROM_PAGE_SIZE), (unsigned char) (address & EEPROM_BYTE_ADDRESS_MASK), get_char() );
					address++;	// Select next eeprom-address.
					i--; // Reduce number of bytes to write by two.
				} while(i); // Loop until all bytes written.
				send_char('\r');
			}
			// support flush
			else if(val=='F' || val=='U')
			{
				address<<=1;
				ADDR_T tempaddress=address;
				do
				{
#ifdef USE_WATCHDOG
					WDT_Reset();
#endif // USE_WATCHDOG
					data = get_char();
					data |= (get_char() << 8);
					SP_LoadFlashWord(address, data);
					address+=2;	// Select next word in memory.
					i -= 2; // Reduce number of bytes to write by two.
				} while(i); // Loop until all bytes written.

				if (val == 'F')
				{
					SP_WriteApplicationPage(tempaddress);
				}
				else if (val == 'U')
				{
					SP_EraseUserSignatureRow();
					SP_WaitForSPM();
					SP_WriteUserSignatureRow();
				}
				SP_WaitForSPM();
				send_char('\r');
				address>>=1;
			}
			else
			{
				send_char('?');
			}
		}
		// Block read
		else if (val == 'g')
		{
			// Block size
			i = (get_char() << 8) | get_char();
			// Memory type
			val = get_char();

			// support eeprom
			if(val=='E')
			{
				do
				{
#ifdef USE_WATCHDOG
					WDT_Reset();
#endif // USE_WATCHDOG
					send_char( EEPROM_ReadByte( (unsigned char)(address / EEPROM_PAGE_SIZE), (unsigned char) (address & EEPROM_BYTE_ADDRESS_MASK) ) );
					address++;
					i--; // Reduce number of bytes to write by two.
				} while(i); // Loop until all bytes written.
			}
			// support flush
			else if(val=='F' || val == 'U' || val == 'P')
			{
				address<<=1;
				// Read it
				do
				{
#ifdef USE_WATCHDOG
					WDT_Reset();
#endif // USE_WATCHDOG
					if (val == 'F')
					{
						send_char( SP_ReadByte( address++) );
						send_char( SP_ReadByte( address++) );
					}
					else if (val == 'U')
					{
						send_char( SP_ReadUserSignatureByte( address++) );
						send_char( SP_ReadUserSignatureByte( address++) );
					}
					else if (val == 'P')
					{
						send_char( SP_ReadCalibrationByte( address++) );
						send_char( SP_ReadCalibrationByte( address++) );
					}

					SP_WaitForSPM();
					i -= 2; // Reduce number of bytes to write by two.
				} while(i); // Loop until all bytes written.
				address>>=1;
			}
			else
			{
				send_char('?');
			}
		}
#endif // ENABLE_BLOCK_SUPPORT
#ifdef ENABLE_FLASH_BYTE_SUPPORT
		// Read program memory byte
		else if (val == 'R')
		{
			SP_WaitForSPM();
			send_char(SP_ReadByte((address << 1)+1));
			send_char(SP_ReadByte((address << 1)+0));

			address++;
		}
		// Write program memory low byte
		else if (val == 'c')
		{
			// get low byte
			i = get_char();
			send_char('\r');
		}
		// Write program memory high byte
		else if (val == 'C')
		{
			// get high byte; combine
			i |= (get_char() << 8);
			SP_WaitForSPM();
			SP_LoadFlashWord((address << 1), i);
			address++;
			send_char('\r');
		}
		// Write page
		else if (val == 'm')
		{
			if (address >= (APP_SECTION_SIZE>>1))
			{
				// don't allow bootloader overwrite
				send_char('?');
			}
			else
			{
				SP_WaitForSPM();
				SP_WriteApplicationPage( address << 1);
				send_char('\r');
			}
		}
#endif // ENABLE_FLASH_BYTE_SUPPORT
#ifdef ENABLE_EEPROM_BYTE_SUPPORT
		// Write EEPROM memory
		else if (val == 'D')
		{
			EEPROM_WriteByte( (unsigned char)(address / EEPROM_PAGE_SIZE), (unsigned char) (address & EEPROM_BYTE_ADDRESS_MASK), get_char() );
			address++;
		}
		// Read EEPROM memory
		else if (val == 'd')
		{
			send_char( EEPROM_ReadByte( (unsigned char)(address / EEPROM_PAGE_SIZE), (unsigned char) (address & EEPROM_BYTE_ADDRESS_MASK) ) );
			address++;
		}
#endif // ENABLE_EEPROM_BYTE_SUPPORT
#ifdef ENABLE_LOCK_BITS
		// Write lockbits
		else if (val == 'l')
		{
			SP_WaitForSPM();
			SP_WriteLockBits( get_char() );
			send_char('\r');
		}
		// Read lockbits
		else if (val == 'r')
		{
			SP_WaitForSPM();
			send_char(SP_ReadLockBits());
		}
#endif // ENABLE_LOCK_BITS
#ifdef ENABLE_FUSE_BITS
		// Read low fuse bits
		else if (val == 'F')
		{
			SP_WaitForSPM();
			send_char(SP_ReadFuseByte(0));
		}
		// Read high fuse bits
		else if (val == 'N')
		{
			SP_WaitForSPM();
			send_char(SP_ReadFuseByte(1));
		}
		// Read extended fuse bits
		else if (val == 'Q')
		{
			SP_WaitForSPM();
			send_char(SP_ReadFuseByte(2));
		}
#endif // ENABLE_FUSE_BITS
		// Enter and leave programming mode
		else if ((val == 'P') || (val == 'L'))
		{
			// just acknowledge
			send_char('\r');
		}
		// Exit bootloader
		else if (val == 'E')
		{
			in_bootloader = 0;
			send_char('\r');
			_delay_ms(200);
		}
		// Get programmer type
		else if (val == 'p')
		{
			// serial
			send_char('S');
		}
		// Return supported device codes
		else if (val == 't')
		{
			// send only this device
			send_char(AVR910_CODE); // TODO
			// terminator
			send_char(0);
		}
		// Set LED, clear LED, and set device type
		else if ((val == 'x') || (val == 'y') || (val == 'T'))
		{
			// discard parameter
			get_char();
			send_char('\r');
		}
		// Return program identifier
		else if (val == 'S')
		{
			send_char('X');
			send_char('B');
			send_char('o');
			send_char('o');
			send_char('t');
			send_char('+');
			send_char('+');
		}
		// Read software version
		else if (val == 'V')
		{
			send_char('1');
			send_char('6');
		}
		// Read signature bytes
		else if (val == 's')
		{
			send_char(SIGNATURE_2);
			send_char(SIGNATURE_1);
			send_char(SIGNATURE_0);
		}
		// ESC (0x1b) to sync
		// otherwise, error
		else if (val != 0x1b)
		{
			send_char('?');
		}
	}
        
	// Wait for any lingering SPM instructions to finish
	SP_WaitForSPM();
        
	// Bootloader exit section
	// Code here runs after the bootloader has exited,
	// but before the application code has started
	// --------------------------------------------------
        
#ifdef LOCK_SPM_ON_EXIT
	// Lock SPM writes
	SP_LockSPM();
#endif // LOCK_SPM_ON_EXIT
        
#ifdef USE_LED
	// Turn off LED on exit
	LED_PORT.DIRCLR = (1 << LED_PIN);
#endif // USE_LED
        
       
#ifdef USE_WATCHDOG
	temp = (WDT.CTRL & ~WDT_ENABLE_bm) | WDT_CEN_bm;
	CCP = CCP_IOREG_gc;
	WDT.CTRL = temp;
	//WDT_Disable();
#endif // USE_WATCHDOG

	// disable USB
	USB.CTRLA = 0;
	USB.CTRLB = 0;
        
	// --------------------------------------------------
	// End bootloader exit section
        
	// Jump into main code
	EIND = 0x00;
	reset_vect();
}

void memc(USB_EP_t *ep)
{
	unsigned char *dst=cdc_rxb.data;
	unsigned char *src=(unsigned char *)ep->DATAPTR;
	unsigned char len=ep->CNT;

	cdc_rxb.len=len;
	while(len)
	{
		*dst=*src;
		dst++;
		src++;
		len--;
	}
	cdc_rxb.bytes=0;
	cdc_rxb.flag=0;
}

unsigned char __attribute__ ((noinline)) get_char(void)
{
	unsigned char ret=0;
        
	while (1)
	{
		// handle usb-Tasks
		USB_T();
//		USB_IO();
		// no data in buffer; wait for new one
		if(cdc_rxb.flag)
		{
			// data ready on usb ?
			if(!epRX_bank)
			{
				// incomming data (out-endpoint)
				// bank ist set ; the next action is made with bank 1
				// if out-endpoint the data from the last aktion (receiving data) are on bank 0
				if(epRX_ptr_out->STATUS&(USB_EP_TRNCOMPL0_bm))
				{
					memc(epRX_ptr_out);
					epRX_ptr_out->CNT=CDC_TXRX_EPSIZE-1;
					LACR16(&(epRX_ptr_out->STATUS), USB_EP_BUSNACK0_bm | USB_EP_TRNCOMPL0_bm | USB_EP_OVF_bm);
					epRX_bank=1;
				}
			}
			else
			{
				// incomming data (out-endpoint)
				// bank ist not set ; the next action is made with bank 0
				// if out-endpoint the data from the last aktion (receiving data) are on bank 1
				if(epRX_ptr_out->STATUS&(USB_EP_TRNCOMPL1_bm))
				{
					memc(epRX_ptr_in);
					epRX_ptr_in->CNT=CDC_TXRX_EPSIZE-1;
					LACR16(&(epRX_ptr_out->STATUS), USB_EP_BUSNACK1_bm | USB_EP_TRNCOMPL1_bm | USB_EP_OVF_bm);
					epRX_bank=0;
				}
			}
		}
		else
		{
			if(cdc_rxb.bytes==cdc_rxb.len)
			{
				// all data processed; send the data in txd and wait for new data
				// only if out-data are processed
				cdc_rxb.flag=1;
			}
			else
			{
				// process data
				ret=cdc_rxb.data[cdc_rxb.bytes];
				cdc_rxb.bytes+=1;
				return ret;
			}
		}
	}
	return ret;
}

void __attribute__ ((noinline)) send_char(unsigned char c)
{
	while(1)
	{
		if(epTX_bank)
		{
			// outgoing data (in-endpoint)
			// bank ist set ; the next action is made with bank 1
			// if in-endpoint the data for the next aktion (sending data) are on bank 1
			if(epTX_ptr_in->STATUS&(USB_EP_BUSNACK1_bm))
			{
				
				*(unsigned char *)epTX_ptr_out->DATAPTR=c;
				epTX_ptr_out->CNT=1;
				LACR16(&(epTX_ptr_in->STATUS), USB_EP_BUSNACK1_bm | USB_EP_TRNCOMPL1_bm);
				epTX_bank=0;
				return;
			}
		}
		else
		{
			// incomming data (out-endpoint)
			// bank ist not set ; the next action is made with bank 0
			// if out-endpoint the data from the last aktion (receiving data) are on bank 1
			if(epTX_ptr_in->STATUS&(USB_EP_BUSNACK0_bm))
			{
				*(unsigned char *)epTX_ptr_in->DATAPTR=c;
				epTX_ptr_in->CNT=1;
				LACR16(&(epTX_ptr_in->STATUS), USB_EP_BUSNACK0_bm | USB_EP_TRNCOMPL0_bm);
				epTX_bank=1;
				return;
			}
		}
	}
}

